import re


def noop(msisdn):
    return msisdn


def vas2nets_voice(msisdn):
    """
    FIXME: this should not need be in this repo

    Vas2Nets is an aggregator in Nigeria, for some reason they need
    MSISDNs prefixed with a 9 instead of the country code to initiate an OBD.
    """
    return re.sub(r'\+?234(\d+)$', r'90\1', msisdn)


def vas2nets_text(msisdn):
    """
    FIXME: this should not need be in this repo

    Vas2Nets is an aggregator in Nigeria, they need MSISDNs in the local
    format, prefixed with a 0, instead of the international format with the
    country code.
    """
    return re.sub(r'\+234(\d+)$', r'234\1', msisdn)



# -*- coding: utf-8 -*-
# Generated by Django 1.9.1 on 2016-11-25 13:38
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('message_sender', '0003_auto_20161124_1357'),
    ]

    operations = [
        migrations.AddField(
            model_name='outbound',
            name='last_sent_time',
            field=models.DateTimeField(blank=True, null=True),
        ),
    ]



"""
Django settings for seed_message_sender project.

For more information on this file, see
https://docs.djangoproject.com/en/1.9/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/1.9/ref/settings/
"""

import os

import dj_database_url

from kombu import Exchange, Queue
import djcelery

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
BASE_DIR = os.path.dirname(os.path.dirname(__file__))

# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/1.9/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = os.environ.get('SECRET_KEY', 'REPLACEME')

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = os.environ.get('DEBUG', False)

ALLOWED_HOSTS = ['*']


# Application definition

INSTALLED_APPS = (
    # admin
    'django.contrib.admin',
    # core
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'django.contrib.sites',
    # documentation
    'rest_framework_docs',
    # 3rd party
    'raven.contrib.django.raven_compat',
    'rest_framework',
    'rest_framework.authtoken',
    'django_filters',
    'rest_hooks',
    'djcelery',
    # us
    'message_sender',

)

SITE_ID = 1
USE_SSL = os.environ.get('USE_SSL', 'false').lower() == 'true'

MIDDLEWARE_CLASSES = (
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.auth.middleware.SessionAuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
)

ROOT_URLCONF = 'seed_message_sender.urls'

WSGI_APPLICATION = 'seed_message_sender.wsgi.application'


# Database
# https://docs.djangoproject.com/en/1.9/ref/settings/#databases

DATABASES = {
    'default': dj_database_url.config(
        default=os.environ.get(
            'MESSAGE_SENDER_DATABASE',
            'postgres://postgres:@localhost/seed_message_sender')),
}


# Internationalization
# https://docs.djangoproject.com/en/1.9/topics/i18n/

LANGUAGE_CODE = 'en-gb'

TIME_ZONE = 'UTC'

USE_I18N = True

USE_L10N = True

USE_TZ = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/1.9/howto/static-files/

STATICFILES_FINDERS = (
    'django.contrib.staticfiles.finders.AppDirectoriesFinder',
    'django.contrib.staticfiles.finders.FileSystemFinder',
)

STATIC_ROOT = 'static'
STATIC_URL = '/static/'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [
            os.path.join(BASE_DIR, 'templates')
        ],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

PAPERTRAIL = os.environ.get('PAPERTRAIL')
if PAPERTRAIL:
    import seed_papertrail  # noqa
    PAPERTRAIL_HOST, _, PAPERTRAIL_PORT = PAPERTRAIL.partition(':')
    LOGGING = seed_papertrail.auto_configure(
        host=PAPERTRAIL_HOST,
        port=int(PAPERTRAIL_PORT),
        system=os.environ.get('MARATHON_APP_DOCKER_IMAGE', 'seed'),
        program=os.environ.get('MESOS_TASK_ID', 'message_sender'))
    LOGGING['loggers']['celery.app'] = {
        'handlers': ['papertrail'],
        'level': 'DEBUG',
        'propagate': True,
    }


# Sentry configuration
RAVEN_CONFIG = {
    # DevOps will supply you with this.
    'dsn': os.environ.get('MESSAGE_SENDER_SENTRY_DSN', None),
}

# REST Framework conf defaults
REST_FRAMEWORK = {
    'PAGE_SIZE': 1000,
    'DEFAULT_PAGINATION_CLASS':
        'rest_framework.pagination.LimitOffsetPagination',
    'DEFAULT_AUTHENTICATION_CLASSES': (
        'rest_framework.authentication.BasicAuthentication',
        'rest_framework.authentication.TokenAuthentication',
    ),
    'DEFAULT_PERMISSION_CLASSES': (
        'rest_framework.permissions.IsAuthenticated',
    ),
    'DEFAULT_FILTER_BACKENDS': ('rest_framework.filters.DjangoFilterBackend',)
}

# Webhook event definition
HOOK_EVENTS = {
    # 'any.event.name': 'App.Model.Action' (created/updated/deleted)
    # 'dummymodel.added': 'message_sender.DummyModel.created+'
}

HOOK_DELIVERER = 'message_sender.tasks.deliver_hook_wrapper'

HOOK_AUTH_TOKEN = os.environ.get('HOOK_AUTH_TOKEN', 'REPLACEME')

# Celery configuration options
CELERY_RESULT_BACKEND = 'djcelery.backends.database:DatabaseBackend'
CELERYBEAT_SCHEDULER = 'djcelery.schedulers.DatabaseScheduler'

BROKER_URL = os.environ.get('BROKER_URL', 'redis://localhost:6379/0')

CELERY_DEFAULT_QUEUE = 'seed_message_sender'
CELERY_QUEUES = (
    Queue('seed_message_sender',
          Exchange('seed_message_sender'),
          routing_key='seed_message_sender'),
)

CELERY_ALWAYS_EAGER = False

# Tell Celery where to find the tasks
CELERY_IMPORTS = (
    'message_sender.tasks',
)

CELERY_CREATE_MISSING_QUEUES = True
CELERY_ROUTES = {
    'celery.backend_cleanup': {
        'queue': 'mediumpriority',
    },
    'message_sender.tasks.deliver_hook_wrapper': {
        'queue': 'priority',
    },
    'message_sender.tasks.send_message': {
        'queue': 'lowpriority',
    },
    'message_sender.tasks.fire_metric': {
        'queue': 'metrics',
    },
    'message_sender.tasks.requeue_failed_tasks': {
        'queue': 'mediumpriority',
    },
}

METRICS_REALTIME = [
    'inbounds.created.sum',
    'vumimessage.tries.sum',
    'vumimessage.maxretries.sum',
    'vumimessage.obd.tries.sum',
    'vumimessage.obd.successful.sum',
    'vumimessage.obd.unsuccessful.sum',
    'message.failures.sum',
    'message.sent.sum',
    'sender.send_message.connection_error.sum',
    'sender.send_message.http_error.400.sum',
    'sender.send_message.http_error.401.sum',
    'sender.send_message.http_error.403.sum',
    'sender.send_message.http_error.404.sum',
    'sender.send_message.http_error.500.sum',
    'sender.send_message.timeout.sum',
]
METRICS_SCHEDULED = [
]
METRICS_SCHEDULED_TASKS = [
]

CELERY_TASK_SERIALIZER = 'json'
CELERY_RESULT_SERIALIZER = 'json'
CELERY_ACCEPT_CONTENT = ['json']
CELERY_IGNORE_RESULT = True
CELERYD_MAX_TASKS_PER_CHILD = 50

djcelery.setup_loader()

MESSAGE_BACKEND_VOICE = os.environ.get(
    'MESSAGE_SENDER_MESSAGE_BACKEND_VOICE', 'vumi')
MESSAGE_BACKEND_TEXT = os.environ.get(
    'MESSAGE_SENDER_MESSAGE_BACKEND_TEXT', 'vumi')

VUMI_API_URL_VOICE = \
    os.environ.get('MESSAGE_SENDER_VUMI_API_URL_VOICE',
                   'http://example.com/api/v1/go/http_api_nostream')
VUMI_ACCOUNT_KEY_VOICE = \
    os.environ.get('MESSAGE_SENDER_VUMI_ACCOUNT_KEY_VOICE', 'acc-key')
VUMI_CONVERSATION_KEY_VOICE = \
    os.environ.get('MESSAGE_SENDER_VUMI_CONVERSATION_KEY_VOICE', 'conv-key')
VUMI_ACCOUNT_TOKEN_VOICE = \
    os.environ.get('MESSAGE_SENDER_VUMI_ACCOUNT_TOKEN_VOICE', 'conv-token')

VOICE_TO_ADDR_FORMATTER = os.environ.get(
    'VOICE_TO_ADDR_FORMATTER', 'message_sender.formatters.noop')
TEXT_TO_ADDR_FORMATTER = os.environ.get(
    'TEXT_TO_ADDR_FORMATTER', 'message_sender.formatters.noop')

VUMI_API_URL_TEXT = \
    os.environ.get('MESSAGE_SENDER_VUMI_API_URL_TEXT',
                   'http://example.com/api/v1/go/http_api_nostream')
VUMI_ACCOUNT_KEY_TEXT = \
    os.environ.get('MESSAGE_SENDER_VUMI_ACCOUNT_KEY_TEXT', 'acc-key')
VUMI_CONVERSATION_KEY_TEXT = \
    os.environ.get('MESSAGE_SENDER_VUMI_CONVERSATION_KEY_TEXT', 'conv-key')
VUMI_ACCOUNT_TOKEN_TEXT = \
    os.environ.get('MESSAGE_SENDER_VUMI_ACCOUNT_TOKEN_TEXT', 'conv-token')

JUNEBUG_API_URL_VOICE = \
    os.environ.get('MESSAGE_SENDER_JUNEBUG_API_URL_VOICE',
                   'http://example.com/jb/channels/abc-def/messages')
JUNEBUG_API_AUTH_VOICE = \
    os.environ.get('MESSAGE_SENDER_JUNEBUG_API_AUTH_VOICE', None)
JUNEBUG_API_FROM_VOICE = \
    os.environ.get('MESSAGE_SENDER_JUNEBUG_API_FROM_VOICE', None)

JUNEBUG_API_URL_TEXT = \
    os.environ.get('MESSAGE_SENDER_JUNEBUG_API_URL_TEXT',
                   'http://example.com/jb/channels/def-abc/messages')
JUNEBUG_API_AUTH_TEXT = \
    os.environ.get('MESSAGE_SENDER_JUNEBUG_API_AUTH_TEXT', None)
JUNEBUG_API_FROM_TEXT = \
    os.environ.get('MESSAGE_SENDER_JUNEBUG_API_FROM_TEXT', None)

MESSAGE_SENDER_MAX_RETRIES = \
    int(os.environ.get('MESSAGE_SENDER_MAX_RETRIES', 3))
MESSAGE_SENDER_MAX_FAILURES = \
    int(os.environ.get('MESSAGE_SENDER_MAX_FAILURES', 5))

METRICS_URL = os.environ.get("METRICS_URL", None)
METRICS_AUTH_TOKEN = os.environ.get("METRICS_AUTH_TOKEN", "REPLACEME")

REDIS_HOST = os.environ.get("REDIS_HOST", "localhost")
REDIS_PORT = os.environ.get("REDIS_PORT", 6379)
REDIS_DB = os.environ.get("REDIS_DB", 0)

# A value of 0 disables cuncurrency limiter
CONCURRENT_VOICE_LIMIT = os.environ.get("CONCURRENT_VOICE_LIMIT", 0)
# Seconds to wait before retrying a waiting message
VOICE_MESSAGE_DELAY = os.environ.get("VOICE_MESSAGE_DELAY", 0)
# Seconds until we assume a message has finished
VOICE_MESSAGE_TIMEOUT = os.environ.get("VOICE_MESSAGE_TIMEOUT", 0)
# A value of 0 disables cuncurrency limiter
CONCURRENT_TEXT_LIMIT = os.environ.get("CONCURRENT_TEXT_LIMIT", 0)
# Seconds to wait before retrying a waiting message
TEXT_MESSAGE_DELAY = os.environ.get("TEXT_MESSAGE_DELAY", 0)
# Seconds until we assume a message has finished
TEXT_MESSAGE_TIMEOUT = os.environ.get("TEXT_MESSAGE_TIMEOUT", 0)

CACHES = {
    'default': {
        'BACKEND': 'redis_cache.RedisCache',
        'LOCATION': ['%s:%s' % (os.environ.get("REDIS_HOST", "localhost"),
                                os.environ.get("REDIS_PORT", 6379))],
        'OPTIONS': {
            'DB': os.environ.get("REDIS_DB", 0),
        }
    },
}
REDIS_PASSWORD = os.environ.get("REDIS_PASSWORD", None)
if REDIS_PASSWORD:
    CACHES['default']['OPTIONS']['PASSWORD'] = REDIS_PASSWORD

DEFAULT_REQUEST_TIMEOUT = float(os.environ.get("DEFAULT_REQUEST_TIMEOUT", 30))



from django.contrib import admin

from .models import Outbound, Inbound


class OutboundAdmin(admin.ModelAdmin):
    list_display = ('to_addr', 'delivered', 'attempts', 'vumi_message_id',
                    'created_at', 'updated_at', 'content', )
    list_filter = ('delivered', 'attempts', 'created_at', 'updated_at', )
    search_fields = ['to_addr']


class InboundAdmin(admin.ModelAdmin):
    list_display = ('message_id', 'in_reply_to', 'to_addr', 'from_addr',
                    'created_at', 'updated_at', 'content', )
    list_filter = ('in_reply_to', 'from_addr', 'created_at', 'updated_at', )
    search_fields = ['to_addr']


admin.site.register(Outbound, OutboundAdmin)
admin.site.register(Inbound, InboundAdmin)



# -*- coding: utf-8 -*-
# Generated by Django 1.9.1 on 2016-12-01 13:53
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('message_sender', '0004_outbound_last_sent_time'),
    ]

    operations = [
        migrations.AddField(
            model_name='outbound',
            name='call_answered',
            field=models.NullBooleanField(default=None, help_text=b'True if the call has been answered. Not used for text messages'),
        ),
    ]



from django.conf import settings
from importlib import import_module


def get_available_metrics():
    available_metrics = []
    available_metrics.extend(settings.METRICS_REALTIME)
    available_metrics.extend(settings.METRICS_SCHEDULED)

    return available_metrics


def load_callable(dotted_path_to_callable):
    module_name, func_name = dotted_path_to_callable.rsplit('.', 1)
    mod = import_module(module_name)
    func = getattr(mod, func_name)
    return func



import json
from go_http.send import LoggingSender

from message_sender.factory import MessageClientFactory
from message_sender.tasks import send_message
from message_sender.tests import AuthenticatedAPITestCase


class TestSendMessage(AuthenticatedAPITestCase):
    """
    Tests related to the send_message task.
    """
    def replace_message_client_factory(self):
        self._patched_message_client_factory_create = (
            MessageClientFactory.create)
        MessageClientFactory.create = staticmethod(
            lambda _: LoggingSender('go_http.test'))

    def restore_message_client_factory(self):
        MessageClientFactory.create = (
            self._patched_message_client_factory_create)

    def test_call_start_end_vumi(self):
        """
        An outbound call should start with us dialling the number, and when
        the person picks up, playing the content and then terminating the call.
        """
        self.replace_message_client_factory()
        outbound = self.make_voice_outbound()
        send_message(outbound.pk)
        outbound.refresh_from_db()

        self.assertTrue(self.check_logs(
            "Message: None sent to '%s' [session_event: new]" % (
                outbound.to_addr)))

        ack = {
            "message_type": "event",
            "event_id": "b04ec322fc1c4819bc3f28e6e0c69de6",
            "event_type": "ack",
            "user_message_id": outbound.vumi_message_id,
            "timestamp": "2015-10-28 16:19:37.485612",
        }
        self.client.post(
            '/api/v1/events', json.dumps(ack), content_type='application/json')

        self.assertTrue(self.check_logs(
            "Message: 'Simple outbound message' sent to '%s' "
            "[session_event: close] "
            "[voice: {'speech_url': 'http://test.com'}]" % (
                outbound.to_addr)))
        self.restore_message_client_factory()

    def test_call_start_end_junebug(self):
        """
        An outbound call should start with us dialling the number, and when
        the person picks up, playing the content and then terminating the call.
        """
        self.replace_message_client_factory()
        outbound = self.make_voice_outbound()
        send_message(outbound.pk)
        outbound.refresh_from_db()

        self.assertTrue(self.check_logs(
            "Message: None sent to '%s' [session_event: new]" % (
                outbound.to_addr)))

        ack = {
            "event_id": "b04ec322fc1c4819bc3f28e6e0c69de6",
            "event_type": "submitted",
            "message_id": outbound.vumi_message_id,
            "timestamp": "2015-10-28 16:19:37.485612",
        }
        self.client.post(
            '/api/v1/events/junebug', json.dumps(ack),
            content_type='application/json')

        self.assertTrue(self.check_logs(
            "Message: 'Simple outbound message' sent to '%s' "
            "[session_event: close] "
            "[voice: {'speech_url': 'http://test.com'}]" % (
                outbound.to_addr)))
        self.restore_message_client_factory()



from django.core.exceptions import ObjectDoesNotExist
from django.conf import settings
from django.contrib.auth.models import User
from rest_hooks.models import Hook
from rest_framework import viewsets, status, filters, mixins
from rest_framework.permissions import IsAuthenticated, AllowAny, IsAdminUser
from rest_framework.views import APIView
from rest_framework.response import Response
from rest_framework.authtoken.models import Token
from requests.exceptions import HTTPError

from .models import Outbound, Inbound, OutboundSendFailure
from .factory import MessageClientFactory
from .serializers import (OutboundSerializer, InboundSerializer,
                          JunebugInboundSerializer, HookSerializer,
                          CreateUserSerializer, OutboundSendFailureSerializer)
from .tasks import (send_message, fire_metric, ConcurrencyLimiter,
                    requeue_failed_tasks)
from seed_message_sender.utils import get_available_metrics, load_callable
import django_filters


voice_to_addr_formatter = load_callable(settings.VOICE_TO_ADDR_FORMATTER)

# Uncomment line below if scheduled metrics are added
# from .tasks import scheduled_metrics


class UserView(APIView):
    """ API endpoint that allows users creation and returns their token.
    Only admin users can do this to avoid permissions escalation.
    """
    permission_classes = (IsAdminUser,)

    def post(self, request):
        '''Create a user and token, given an email. If user exists just
        provide the token.'''
        serializer = CreateUserSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        email = serializer.validated_data.get('email')
        try:
            user = User.objects.get(username=email)
        except User.DoesNotExist:
            user = User.objects.create_user(email, email=email)
        token, created = Token.objects.get_or_create(user=user)

        return Response(
            status=status.HTTP_201_CREATED, data={'token': token.key})


class HookViewSet(viewsets.ModelViewSet):
    """
    Retrieve, create, update or destroy webhooks.
    """
    permission_classes = (IsAuthenticated,)
    queryset = Hook.objects.all()
    serializer_class = HookSerializer

    def perform_create(self, serializer):
        serializer.save(user=self.request.user)


class OutboundFilter(filters.FilterSet):
    before = django_filters.IsoDateTimeFilter(name="created_at",
                                              lookup_type='lte')
    after = django_filters.IsoDateTimeFilter(name="created_at",
                                             lookup_type='gte')

    class Meta:
        model = Outbound
        fields = ('version', 'to_addr', 'vumi_message_id',
                  'delivered', 'attempts', 'metadata',
                  'created_at', 'updated_at',
                  'before', 'after')


class OutboundViewSet(viewsets.ModelViewSet):
    """
    API endpoint that allows Outbound models to be viewed or edited.
    """
    permission_classes = (IsAuthenticated,)
    queryset = Outbound.objects.all()
    serializer_class = OutboundSerializer
    filter_class = OutboundFilter


class InboundViewSet(viewsets.ModelViewSet):

    """
    API endpoint that allows Inbound models to be viewed or edited.
    """
    permission_classes = (IsAuthenticated,)
    queryset = Inbound.objects.all()
    filter_fields = ('message_id', 'in_reply_to', 'to_addr', 'from_addr',
                     'content', 'transport_name', 'transport_type',
                     'helper_metadata', 'created_at', 'updated_at',)

    def get_serializer_class(self):
        if self.action == 'create':
            if "channel_data" in self.request.data:
                return JunebugInboundSerializer
        return InboundSerializer

    def create(self, request, *args, **kwargs):
        close_event = False
        if "channel_data" in request.data:  # Handle message from Junebug
            if "session_event" in request.data["channel_data"]:
                if request.data["channel_data"]["session_event"] == "close":
                    close_event = True
                    reply_field = "reply_to"
                    from_field = "from"
        elif "session_event" in request.data:  # Handle message from Vumi
            if request.data["session_event"] == "close":
                close_event = True
                reply_field = "in_reply_to"
                from_field = "from_addr"

        if close_event:
            try:
                message = Outbound.objects.get(
                    vumi_message_id=request.data[reply_field])
            except ObjectDoesNotExist:
                message = Outbound.objects.filter(
                    to_addr=request.data[from_field]).order_by(
                    '-created_at').last()
            if message:
                outbound_type = "voice" if "voice_speech_url" in \
                    message.metadata else "text"
                ConcurrencyLimiter.decr_message_count(
                    outbound_type, message.last_sent_time)

        return super(InboundViewSet, self).create(request, *args, **kwargs)


class EventListener(APIView):

    """
    Triggers updates to outbound messages based on event data from Vumi
    """
    permission_classes = (AllowAny,)

    def _handle_voice_ack(self, message, timestamp):
        if message.call_answered:
            message.delivered = True
            message.metadata["ack_timestamp"] = timestamp
            message.save(update_fields=['delivered', 'metadata'])
            fire_metric.apply_async(kwargs={
                "metric_name":
                    'vumimessage.obd.successful.sum',
                "metric_value": 1.0
            })
        else:
            message.call_answered = True
            message.metadata["call_answered_timestamp"] = timestamp

            client = MessageClientFactory.create('voice')
            speech_url = message.metadata['voice_speech_url']
            try:
                vumiresponse = client.send_voice(
                    voice_to_addr_formatter(message.to_addr), message.content,
                    speech_url=speech_url, session_event="close")
            except HTTPError:
                message.save(update_fields=['metadata'])
                send_message.delay(message_id=str(message.id))
                return

            message.vumi_message_id = vumiresponse["message_id"]
            message.save(update_fields=[
                'call_answered', 'metadata', 'vumi_message_id'])

    def _handle_text_ack(self, message, timestamp):
        message.delivered = True
        message.metadata["ack_timestamp"] = timestamp
        message.save(update_fields=['delivered', 'metadata'])

    def post(self, request, *args, **kwargs):
        """
        Checks for expect event types before continuing
        """

        try:
            expect = ["message_type", "event_type", "user_message_id",
                      "event_id", "timestamp"]
            if set(expect).issubset(request.data.keys()):
                # Load message
                message = Outbound.objects.get(
                    vumi_message_id=request.data["user_message_id"])
                # only expecting `event` on this endpoint
                if request.data["message_type"] == "event":
                    event = request.data["event_type"]
                    # expecting ack, nack, delivery_report
                    if event == "ack":
                        if "voice_speech_url" in message.metadata:
                            self._handle_voice_ack(
                                message, request.data['timestamp'])
                        else:
                            self._handle_text_ack(
                                message, request.data['timestamp'])
                    elif event == "delivery_report":
                        message.delivered = True
                        message.metadata["delivery_timestamp"] = \
                            request.data["timestamp"]
                        message.save()
                    elif event == "nack":
                        if "nack_reason" in request.data:
                            message.metadata["nack_reason"] = \
                                request.data["nack_reason"]
                            message.save()
                        send_message.delay(str(message.id))
                        if "voice_speech_url" in message.metadata:
                            fire_metric.apply_async(kwargs={
                                "metric_name":
                                    'vumimessage.obd.unsuccessful.sum',
                                "metric_value": 1.0
                            })

                    # Return
                    status = 200
                    accepted = {"accepted": True}
                else:
                    status = 400
                    accepted = {"accepted": False,
                                "reason": "Unexpected message_type"}
            else:
                status = 400
                accepted = {"accepted": False,
                            "reason": "Missing expected body keys"}
        except ObjectDoesNotExist:
            status = 400
            accepted = {"accepted": False,
                        "reason": "Missing message in control"}
        return Response(accepted, status=status)

    # TODO make this work in test harness, works in production
    # def perform_create(self, serializer):
    #     serializer.save(created_by=self.request.user,
    #                     updated_by=self.request.user)
    #
    # def perform_update(self, serializer):
    #     serializer.save(updated_by=self.request.user)


class JunebugEventListener(EventListener):

    """
    Triggers updates to outbound messages based on event data from Junebug
    """
    permission_classes = (AllowAny,)

    def post(self, request, *args, **kwargs):
        """
        Updates the message from the event data.
        """
        expect = ["event_type", "message_id", "timestamp"]
        if not set(expect).issubset(request.data.keys()):
            return Response({
                "accepted": False,
                "reason": "Missing expected body keys"
            }, status=400)

        try:
            message = Outbound.objects.get(
                vumi_message_id=request.data["message_id"])
        except ObjectDoesNotExist:
            return Response({
                "accepted": False,
                "reason": "Cannot find message for event"
            }, status=400)

        event_type = request.data["event_type"]
        if event_type == "submitted":
            if "voice_speech_url" in message.metadata:
                self._handle_voice_ack(message, request.data["timestamp"])
            else:
                self._handle_text_ack(message, request.data["timestamp"])
        elif event_type == "rejected":
            message.metadata["nack_reason"] = (
                request.data.get("event_details"))
            message.save(update_fields=['metadata'])
            send_message.delay(str(message.id))
        elif event_type == "delivery_succeeded":
            message.delivered = True
            message.metadata["delivery_timestamp"] = request.data["timestamp"]
            message.save(update_fields=['delivered', 'metadata'])
        elif event_type == "delivery_failed":
            message.metadata["delivery_failed_reason"] = (
                request.data.get("event_details"))
            message.save(update_fields=['metadata'])
            send_message.delay(str(message.id))

        if ("voice_speech_url" in message.metadata and
                event_type in ("rejected", "delivery_failed")):
            fire_metric.apply_async(kwargs={
                "metric_name": 'vumimessage.obd.unsuccessful.sum',
                "metric_value": 1.0
            })

        return Response({"accepted": True}, status=200)


class MetricsView(APIView):

    """ Metrics Interaction
        GET - returns list of all available metrics on the service
        POST - starts up the task that fires all the scheduled metrics
    """
    permission_classes = (IsAuthenticated,)

    def get(self, request, *args, **kwargs):
        status = 200
        resp = {
            "metrics_available": get_available_metrics()
        }
        return Response(resp, status=status)

    def post(self, request, *args, **kwargs):
        status = 201
        # Uncomment line below if scheduled metrics are added
        # scheduled_metrics.apply_async()
        resp = {"scheduled_metrics_initiated": True}
        return Response(resp, status=status)


class HealthcheckView(APIView):

    """ Healthcheck Interaction
        GET - returns service up - getting auth'd requires DB
    """
    permission_classes = (IsAuthenticated,)

    def get(self, request, *args, **kwargs):
        import seed_message_sender
        import django
        import rest_framework
        status = 200
        resp = {
            "up": True,
            "result": {
                "database": "Accessible",
                "version": seed_message_sender.__version__,
                "libraries": {
                    "django": django.__version__,
                    "djangorestframework": rest_framework.__version__
                }
            }
        }
        return Response(resp, status=status)


class FailedTaskViewSet(mixins.ListModelMixin,
                        mixins.RetrieveModelMixin,
                        viewsets.GenericViewSet):
    permission_classes = (IsAuthenticated,)
    queryset = OutboundSendFailure.objects.all()
    serializer_class = OutboundSendFailureSerializer

    def create(self, request):
        status = 201
        resp = {'requeued_failed_tasks': True}
        requeue_failed_tasks.delay()
        return Response(resp, status=status)



# -*- coding: utf-8 -*-
# Generated by Django 1.9.12 on 2017-01-19 12:05
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('message_sender', '0005_outbound_call_answered'),
    ]

    operations = [
        migrations.AlterField(
            model_name='outbound',
            name='created_at',
            field=models.DateTimeField(auto_now_add=True, db_index=True),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.12 on 2017-02-15 07:17
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('message_sender', '0006_add_created_at_index'),
    ]

    operations = [
        migrations.CreateModel(
            name='OutboundSendFailure',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('task_id', models.UUIDField()),
                ('initiated_at', models.DateTimeField()),
                ('reason', models.TextField()),
                ('outbound', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='message_sender.Outbound')),
            ],
        ),
    ]



from django.conf import settings
from django.contrib.sites.shortcuts import get_current_site

try:
    from urlparse import urlunparse
except ImportError:
    from urllib.parse import urlunparse


def make_absolute_url(path):
    # NOTE: We're using the default site as set by
    #       settings.SITE_ID and the Sites framework
    site = get_current_site(None)
    return urlunparse(
        ('https' if settings.USE_SSL else 'http',
         site.domain, path,
         '', '', ''))



import uuid

from django.contrib.postgres.fields import JSONField
from django.contrib.auth.models import User
from django.db import models
from django.utils.encoding import python_2_unicode_compatible


@python_2_unicode_compatible
class Outbound(models.Model):

    """
    Contacts outbound messages and their status.
    Delivered is set to true when ack received because delivery reports patchy
    """
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    to_addr = models.CharField(null=False, blank=False, max_length=500)
    version = models.IntegerField(default=1)
    content = models.CharField(null=True, blank=True, max_length=1000)
    vumi_message_id = models.CharField(null=True, blank=True, max_length=36,
                                       db_index=True)
    delivered = models.BooleanField(default=False)
    call_answered = models.NullBooleanField(
        default=None, null=True, blank=True, help_text="True if the call has "
        "been answered. Not used for text messages")
    attempts = models.IntegerField(default=0)
    metadata = JSONField()
    updated_at = models.DateTimeField(auto_now=True)
    created_at = models.DateTimeField(auto_now_add=True, db_index=True)
    last_sent_time = models.DateTimeField(null=True, blank=True)
    created_by = models.ForeignKey(User, related_name='outbounds_created',
                                   null=True)
    updated_by = models.ForeignKey(User, related_name='outbounds_updated',
                                   null=True)
    user = property(lambda self: self.created_by)

    def __str__(self):  # __unicode__ on Python 2
        return str(self.id)


@python_2_unicode_compatible
class Inbound(models.Model):

    """
    Contacts inbound messages from Vumi
    """
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    message_id = models.CharField(null=False, blank=False, max_length=36)
    in_reply_to = models.CharField(null=True, blank=True, max_length=36)
    to_addr = models.CharField(null=False, blank=False, max_length=255)
    from_addr = models.CharField(null=False, blank=False, max_length=255)
    content = models.CharField(null=True, blank=True, max_length=1000)
    transport_name = models.CharField(null=False, blank=False, max_length=200)
    transport_type = models.CharField(null=True, blank=True, max_length=200)
    helper_metadata = JSONField()
    updated_at = models.DateTimeField(auto_now=True)
    created_at = models.DateTimeField(auto_now_add=True)
    created_by = models.ForeignKey(User, related_name='inbounds_created',
                                   null=True)
    updated_by = models.ForeignKey(User, related_name='inbounds_updated',
                                   null=True)
    user = property(lambda self: self.created_by)

    def __str__(self):  # __unicode__ on Python 2
        return str(self.id)


@python_2_unicode_compatible
class OutboundSendFailure(models.Model):
    outbound = models.ForeignKey(Outbound, on_delete=models.CASCADE)
    task_id = models.UUIDField()
    initiated_at = models.DateTimeField()
    reason = models.TextField()

    def __str__(self):  # __unicode__ on Python 2
        return str(self.id)



from django.conf.urls import url, include
from rest_framework import routers
from . import views

router = routers.DefaultRouter()
router.register(r'inbound', views.InboundViewSet)
router.register(r'outbound', views.OutboundViewSet)
router.register(r'webhook', views.HookViewSet)
router.register(r'failed-tasks', views.FailedTaskViewSet)

# Wire up our API using automatic URL routing.
# Additionally, we include login URLs for the browseable API.
urlpatterns = [
    url('^api/v1/events$',
        views.EventListener.as_view()),
    url(
        '^api/v1/events/junebug$', views.JunebugEventListener.as_view(),
        name='junebug-events'),
    url(r'^api/v1/user/token/$', views.UserView.as_view(),
        name='create-user-token'),
    url(r'^api/v1/', include(router.urls)),
]



import json

from django.conf import settings
from django.core.urlresolvers import reverse

import requests

from go_http.send import HttpApiSender

from .utils import make_absolute_url


class FactoryException(Exception):
    pass


def get_backend_type(client_type):
    backend_type = getattr(
        settings, 'MESSAGE_BACKEND_%s' % client_type.upper(), None)
    if not backend_type:
        raise FactoryException(
            'Undefined message backend: %r' % (backend_type,))
    return backend_type.lower()


class JunebugApiSenderException(Exception):
    pass


class JunebugApiSender(HttpApiSender):

    def __init__(self, url, auth=None, from_addr=None, session=None):
        """
        :param url str: The URL for the Junebug HTTP channel
        :param auth tuple: (username, password) or anything
            accepted by the requests library. Defaults to None.
        :param session requests.Session: A requests session. Defaults to None
        :param from_addr str: The from address for all messages. Defaults to
            None
        """
        self.api_url = url
        self.auth = auth
        self.from_addr = from_addr
        if session is None:
            session = requests.Session()
        self.session = session

    def _raw_send(self, py_data):
        headers = {'content-type': 'application/json; charset=utf-8'}

        channel_data = py_data.get('helper_metadata', {})
        channel_data['session_event'] = py_data.get('session_event')

        data = {
            'to': py_data['to_addr'],
            'from': self.from_addr,
            'content': py_data['content'],
            'channel_data': channel_data,
            'event_url': make_absolute_url(reverse('junebug-events')),
        }

        data = json.dumps(data)
        r = self.session.post(self.api_url, auth=self.auth,
                              data=data, headers=headers,
                              timeout=settings.DEFAULT_REQUEST_TIMEOUT)
        r.raise_for_status()
        res = r.json()
        return res.get('result', {})

    def fire_metric(self, metric, value, agg="last"):
        raise JunebugApiSenderException(
            'Metrics sending not supported by Junebug')


class MessageClientFactory(object):

    @classmethod
    def create(cls, client_type):
        backend_type = get_backend_type(client_type)
        handler = getattr(cls,
                          'create_%s_client' % (backend_type,), None)
        if not handler:
            raise FactoryException(
                'Unknown backend type: %r' % (backend_type,))

        return handler(client_type)

    @classmethod
    def create_junebug_client(cls, client_type):
        return JunebugApiSender(
            getattr(settings, 'JUNEBUG_API_URL_%s' % (client_type.upper(),)),
            getattr(settings, 'JUNEBUG_API_AUTH_%s' % (client_type.upper(),)),
            getattr(settings, 'JUNEBUG_API_FROM_%s' % (client_type.upper())))

    @classmethod
    def create_vumi_client(cls, client_type):
        return HttpApiSender(
            getattr(settings,
                    'VUMI_ACCOUNT_KEY_%s' % (client_type.upper(),)),
            getattr(settings,
                    'VUMI_CONVERSATION_KEY_%s' % (client_type.upper(),)),
            getattr(settings,
                    'VUMI_ACCOUNT_TOKEN_%s' % (client_type.upper(),)),
            api_url=getattr(settings,
                            'VUMI_API_URL_%s' % (client_type.upper(),)),
        )



from django.test import TestCase, override_settings

from .utils import make_absolute_url


class TestMakeAbsoluteUrl(TestCase):
    @override_settings(USE_SSL=True)
    def test_make_absolute_url_ssl_true(self):
        """
        If USE_SSL is True, then then the url should be using https
        """
        self.assertEqual(
            make_absolute_url('foo'),
            'https://example.com/foo')
        self.assertEqual(
            make_absolute_url('/foo'),
            'https://example.com/foo')

    @override_settings(USE_SSL=False)
    def test_make_absolute_url_ssl_false(self):
        """
        If USE_SSL is False, then then the url should be using http
        """
        self.assertEqual(
            make_absolute_url('foo'),
            'http://example.com/foo')
        self.assertEqual(
            make_absolute_url('/foo'),
            'http://example.com/foo')




def psh_fire_msg_action_if_new(sender, instance, created, **kwargs):
    """ Post save hook to fire message send task
    """
    if created:
        from message_sender.tasks import send_message
        send_message.apply_async(kwargs={"message_id": str(instance.id)})


def psh_fire_metrics_if_new(sender, instance, created, **kwargs):
    """ Post save hook to fire Inbound created metric
    """
    if created:
        from message_sender.tasks import fire_metric
        fire_metric.apply_async(kwargs={
            "metric_name": 'inbounds.created.sum',
            "metric_value": 1.0
        })



from .models import Inbound, Outbound, OutboundSendFailure
from rest_hooks.models import Hook
from rest_framework import serializers


class OutboundSerializer(serializers.HyperlinkedModelSerializer):

    class Meta:
        model = Outbound
        fields = (
            'url', 'id', 'version', 'to_addr', 'vumi_message_id', 'content',
            'delivered', 'attempts', 'metadata', 'created_at', 'updated_at')


class InboundSerializer(serializers.HyperlinkedModelSerializer):

    class Meta:
        model = Inbound
        fields = (
            'url', 'id', 'message_id', 'in_reply_to', 'to_addr',
            'from_addr', 'content', 'transport_name', 'transport_type',
            'helper_metadata', 'created_at', 'updated_at')

    def to_internal_value(self, data):
        """
        Adds extra data to the helper_metadata field.
        """
        if "session_event" in data:
            data['helper_metadata']['session_event'] = data['session_event']
        return super(InboundSerializer, self).to_internal_value(data)


class JunebugInboundSerializer(serializers.HyperlinkedModelSerializer):
    """
    Maps fields from Junebug onto fields expected by the Inbound model.
    """
    reply_to = serializers.CharField(source='in_reply_to')
    to = serializers.CharField(source='to_addr')
    channel_id = serializers.CharField(source='transport_name')
    channel_data = serializers.JSONField(source='helper_metadata')

    class Meta:
        model = Inbound
        fields = (
            'url', 'id', 'message_id', 'reply_to', 'to', 'from_addr',
            'content', 'channel_id', 'channel_data', 'created_at',
            'updated_at')

    def to_internal_value(self, data):
        """
        Maps Junebug 'from' field to 'from_addr' expected by serializer since
        'from' is a python keyword.
        """
        if "from" not in data:
            raise serializers.ValidationError({
                'from': 'This field is required.'
            })
        else:
            data['from_addr'] = data['from']
        return super(JunebugInboundSerializer, self).to_internal_value(data)


class HookSerializer(serializers.ModelSerializer):

    class Meta:
        model = Hook
        read_only_fields = ('user',)


class CreateUserSerializer(serializers.Serializer):
    email = serializers.EmailField()


class OutboundSendFailureSerializer(serializers.HyperlinkedModelSerializer):

    class Meta:
        model = OutboundSendFailure
        fields = ('url', 'id', 'outbound', 'task_id', 'initiated_at', 'reason')



import json
import random
import requests
import time

from celery.task import Task
from celery.utils.log import get_task_logger

from datetime import datetime
from django.conf import settings
from django.core.cache import cache
from django.core.exceptions import ObjectDoesNotExist

from go_http.metrics import MetricsApiClient
from requests import exceptions as requests_exceptions

from .factory import MessageClientFactory


from .models import Outbound, OutboundSendFailure
from seed_message_sender.utils import load_callable
from seed_papertrail.decorators import papertrail

logger = get_task_logger(__name__)

voice_to_addr_formatter = load_callable(settings.VOICE_TO_ADDR_FORMATTER)
text_to_addr_formatter = load_callable(settings.TEXT_TO_ADDR_FORMATTER)


def calculate_retry_delay(attempt, max_delay=300):
    """Calculates an exponential backoff for retry attempts with a small
    amount of jitter."""
    delay = int(random.uniform(2, 4) ** attempt)
    if delay > max_delay:
        # After reaching the max delay, stop using expontential growth
        # and keep the delay nearby the max.
        delay = int(random.uniform(max_delay - 20, max_delay + 20))
    return delay


class DeliverHook(Task):
    def run(self, target, payload, instance_id=None, hook_id=None, **kwargs):
        """
        target:     the url to receive the payload.
        payload:    a python primitive data structure
        instance_id:   a possibly None "trigger" instance ID
        hook_id:       the ID of defining Hook object
        """
        requests.post(
            url=target,
            data=json.dumps(payload),
            headers={
                'Content-Type': 'application/json',
                'Authorization': 'Token %s' % settings.HOOK_AUTH_TOKEN
            }
        )


def deliver_hook_wrapper(target, payload, instance, hook):
    if instance is not None:
        instance_id = instance.id
    else:
        instance_id = None
    kwargs = dict(target=target, payload=payload,
                  instance_id=instance_id, hook_id=hook.id)
    DeliverHook.apply_async(kwargs=kwargs)


def get_metric_client(session=None):
    return MetricsApiClient(
        auth_token=settings.METRICS_AUTH_TOKEN,
        api_url=settings.METRICS_URL,
        session=session)


class FireMetric(Task):

    """ Fires a metric using the MetricsApiClient
    """
    name = "message_sender.tasks.fire_metric"

    @papertrail.debug(name, sample=0.1)
    def run(self, metric_name, metric_value, session=None, **kwargs):
        metric_value = float(metric_value)
        metric = {
            metric_name: metric_value
        }
        metric_client = get_metric_client(session=session)
        metric_client.fire(metric)
        return "Fired metric <%s> with value <%s>" % (
            metric_name, metric_value)


fire_metric = FireMetric()


class ConcurrencyLimiter(object):
    BUCKET_SIZE = 60

    @classmethod
    def get_key(cls, msg_type, bucket):
        return "%s_messages_at_%s" % (msg_type, bucket)

    @classmethod
    def get_current_message_count(cls, msg_type, timeout):
        # Sum the values in all the buckets to get the total
        total = 0
        number_of_buckets = timeout // cls.BUCKET_SIZE + 1
        bucket = int(time.time() // cls.BUCKET_SIZE)
        for i in range(bucket, bucket - number_of_buckets, -1):
            value = cache.get(cls.get_key(msg_type, i))
            if value:
                total += int(value)
        return total

    @classmethod
    def incr_message_count(cls, msg_type, timeout):
        bucket = int(time.time() // cls.BUCKET_SIZE)
        key = cls.get_key(msg_type, bucket)

        # Add the bucket size to the expiry time so messages that start at
        # the end of the bucket still complete
        if not cache.add(key, 1, timeout + cls.BUCKET_SIZE):
            cache.incr(key)

    @classmethod
    def decr_message_count(cls, msg_type, msg_time):

        if msg_type == "voice":
            if int(getattr(settings, 'CONCURRENT_VOICE_LIMIT', 0)) == 0:
                return
            timeout = int(getattr(settings, 'VOICE_MESSAGE_TIMEOUT', 0))
        else:
            if int(getattr(settings, 'CONCURRENT_TEXT_LIMIT', 0)) == 0:
                return
            timeout = int(getattr(settings, 'TEXT_MESSAGE_TIMEOUT', 0))

        if not msg_time:
            return

        # Convert from datetime to seconds since epoch
        msg_time = (msg_time - datetime(1970, 1, 1)).total_seconds()

        time_since = time.time() - msg_time
        if time_since > timeout:
            return
        bucket = int(msg_time // cls.BUCKET_SIZE)

        key = cls.get_key(msg_type, bucket)
        # Set the expiry time to the timeout minus the time passed since
        # the message was sent.
        if int(cache.get_or_set(key, 0, timeout - time_since)) > 0:
            cache.decr(key)

    @classmethod
    def manage_limit(cls, task, msg_type, limit, timeout, delay):
        if limit > 0:
            if cls.get_current_message_count(msg_type, timeout) >= limit:
                task.retry(countdown=delay)
            cls.incr_message_count(msg_type, timeout)


class SendMessage(Task):

    """
    Task to load and contruct message and send them off
    """
    name = "message_sender.tasks.send_message"
    default_retry_delay = 5
    max_retries = 5

    class FailedEventRequest(Exception):

        """
        The attempted task failed because of a non-200 HTTP return
        code.
        """

    def get_text_client(self):
        return MessageClientFactory.create('text')

    def get_voice_client(self):
        return MessageClientFactory.create('voice')

    @papertrail.debug(name, sample=0.1)
    def run(self, message_id, **kwargs):
        """
        Load and contruct message and send them off
        """
        l = self.get_logger(**kwargs)

        l.info("Loading Outbound Message <%s>" % message_id)
        try:
            message = Outbound.objects.get(id=message_id)
        except ObjectDoesNotExist:
            logger.error('Missing Outbound message', exc_info=True)
            return

        if message.attempts < settings.MESSAGE_SENDER_MAX_RETRIES:
            if self.request.retries > 0:
                retry_delay = calculate_retry_delay(self.request.retries)
            else:
                retry_delay = self.default_retry_delay
            l.info("Attempts: %s" % message.attempts)
            # send or resend
            try:
                if "voice_speech_url" in message.metadata:
                    # OBD number of tries metric
                    fire_metric.apply_async(kwargs={
                        "metric_name": 'vumimessage.obd.tries.sum',
                        "metric_value": 1.0
                    })

                    # Voice message
                    ConcurrencyLimiter.manage_limit(
                        self, "voice",
                        int(getattr(settings,
                                    'CONCURRENT_VOICE_LIMIT', 0)),
                        int(getattr(settings, 'VOICE_MESSAGE_TIMEOUT', 0)),
                        int(getattr(settings, 'VOICE_MESSAGE_DELAY', 0)))
                    sender = self.get_voice_client()
                    # Start call. We send the voice message on the ack.
                    vumiresponse = sender.send_voice(
                        voice_to_addr_formatter(message.to_addr),
                        None, session_event="new")
                    message.call_answered = False
                    l.info("Sent voice message to <%s>" % message.to_addr)
                else:
                    # Plain content
                    ConcurrencyLimiter.manage_limit(
                        self, "text",
                        int(getattr(settings, 'CONCURRENT_TEXT_LIMIT', 0)),
                        int(getattr(settings, 'TEXT_MESSAGE_TIMEOUT', 0)),
                        int(getattr(settings, 'TEXT_MESSAGE_DELAY', 0)))
                    sender = self.get_text_client()
                    vumiresponse = sender.send_text(
                        text_to_addr_formatter(message.to_addr),
                        message.content,
                        session_event="new")
                    l.info("Sent text message to <%s>" % (
                        message.to_addr,))

                message.last_sent_time = datetime.now()
                message.attempts += 1
                message.vumi_message_id = vumiresponse["message_id"]
                message.save()
                fire_metric.apply_async(kwargs={
                    "metric_name": 'vumimessage.tries.sum',
                    "metric_value": 1.0
                })
            except requests_exceptions.ConnectionError as exc:
                l.info('Connection Error sending message')
                fire_metric.delay(
                    'sender.send_message.connection_error.sum', 1)
                self.retry(exc=exc, countdown=retry_delay)
            except requests_exceptions.Timeout as exc:
                l.info('Sending message failed due to timeout')
                fire_metric.delay('sender.send_message.timeout.sum', 1)
                self.retry(exc=exc, countdown=retry_delay)
            except requests_exceptions.HTTPError as exc:
                # retry message sending if in 500 range (3 default
                # retries)
                l.info('Sending message failed due to status: %s' %
                       exc.response.status_code)
                metric_name = ('sender.send_message.http_error.%s.sum' %
                               exc.response.status_code)
                fire_metric.delay(metric_name, 1)
                self.retry(exc=exc, countdown=retry_delay)

            # If we've gotten this far the message send was successful.
            fire_metric.apply_async(kwargs={
                "metric_name": 'message.sent.sum',
                "metric_value": 1.0
            })
            return vumiresponse

        else:
            # This is for retries based on async nacks from the transport.
            l.info("Message <%s> at max retries." % str(message_id))
            fire_metric.apply_async(kwargs={
                "metric_name": 'vumimessage.maxretries.sum',
                "metric_value": 1.0
            })
            # Count failures on exhausted tries.
            fire_metric.apply_async(kwargs={
                "metric_name": 'message.failures.sum',
                "metric_value": 1.0
            })

    def on_failure(self, exc, task_id, args, kwargs, einfo):
        if self.request.retries == self.max_retries:
            OutboundSendFailure.objects.create(
                subscription_id=args[0],
                initiated_at=self.request.eta,
                reason=einfo.exception.message,
                task_id=task_id
            )
            # Count permanent failures.
            fire_metric.apply_async(kwargs={
                "metric_name": 'message.failures.sum',
                "metric_value": 1.0
            })
        super(SendMessage, self).on_failure(exc, task_id, args,
                                            kwargs, einfo)


send_message = SendMessage()


class RequeueFailedTasks(Task):

    """
    Task to requeue failed Outbounds.
    """
    name = "message_sender.tasks.requeue_failed_tasks"

    def run(self, **kwargs):
        l = self.get_logger(**kwargs)
        failures = OutboundSendFailure.objects
        l.info("Attempting to requeue <%s> failed Outbound sends" %
               failures.all().count())
        for failure in failures.iterator():
            outbound_id = str(failure.outbound_id)
            # Cleanup the failure before requeueing it.
            failure.delete()
            send_message.delay(outbound_id)


requeue_failed_tasks = RequeueFailedTasks()



# -*- coding: utf-8 -*-
# Generated by Django 1.9.1 on 2016-11-24 11:50
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('message_sender', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='outbound',
            name='vumi_message_id',
            field=models.CharField(blank=True, db_index=True, max_length=36, null=True),
        ),
    ]



from django.apps import AppConfig

from django.db.models.signals import post_save, pre_save  # noqa


class MessageSenderAppConfig(AppConfig):

    name = 'message_sender'

    def ready(self):
        from .signals import (psh_fire_msg_action_if_new,
                              psh_fire_metrics_if_new)

        post_save.connect(
            psh_fire_msg_action_if_new,
            sender='message_sender.Outbound',
            dispatch_uid='psh_fire_msg_action_if_new')

        post_save.connect(
            psh_fire_metrics_if_new,
            sender='message_sender.Inbound',
            dispatch_uid='psh_fire_metrics_if_new')



import json
import uuid
import logging
import responses

try:
    from urllib.parse import urlparse, urlencode
except ImportError:
    from urlparse import urlparse
    from urllib import urlencode

from datetime import timedelta

from celery.exceptions import Retry
from datetime import datetime
from django.test import TestCase, override_settings
from django.contrib.auth.models import User
from django.db.models.signals import post_save
from django.conf import settings
from django.utils import timezone
from mock import MagicMock
from mock import patch
from rest_framework import status
from rest_framework.test import APIClient
from rest_framework.authtoken.models import Token
from requests_testadapter import TestAdapter, TestSession
from go_http.metrics import MetricsApiClient
from go_http.send import LoggingSender

from .factory import (
    MessageClientFactory, JunebugApiSender, HttpApiSender,
    JunebugApiSenderException, FactoryException)
from .models import Inbound, Outbound, OutboundSendFailure
from .signals import psh_fire_metrics_if_new, psh_fire_msg_action_if_new
from .tasks import (SendMessage, send_message, fire_metric,
                    ConcurrencyLimiter, requeue_failed_tasks)
from . import tasks

from seed_message_sender.utils import load_callable

SendMessage.get_text_client = lambda x: LoggingSender('go_http.test')
SendMessage.get_voice_client = lambda x: LoggingSender('go_http.test')


class RecordingAdapter(TestAdapter):

    """ Record the request that was handled by the adapter.
    """
    request = None

    def send(self, request, *args, **kw):
        self.request = request
        return super(RecordingAdapter, self).send(request, *args, **kw)


class RecordingHandler(logging.Handler):

    """ Record logs. """
    logs = None

    def emit(self, record):
        if self.logs is None:
            self.logs = []

        self.logs.append(record)


class MockCache(object):
    def __init__(self):
        self.cache_data = {}

    def get(self, key):
        return self.cache_data.get(key, None)

    def get_or_set(self, key, value, expire=0):
        if key not in self.cache_data:
            self.cache_data[key] = value
            return value
        return self.cache_data[key]

    def add(self, key, value, expire=0):
        if key not in self.cache_data:
            self.cache_data[key] = value
            return True
        return False

    def incr(self, key, value=1):
        if key not in self.cache_data:
            raise(ValueError)
        self.cache_data[key] += value

    def decr(self, key, value=1):
        if key not in self.cache_data:
            raise(ValueError)
        self.cache_data[key] -= value


class APITestCase(TestCase):

    def setUp(self):
        self.client = APIClient()
        self.adminclient = APIClient()
        self.session = TestSession()


class AuthenticatedAPITestCase(APITestCase):

    def make_outbound(self):
        self._replace_post_save_hooks_outbound()  # don't let fixtures fire
        outbound_message = {
            "to_addr": "+27123",
            "vumi_message_id": "075a32da-e1e4-4424-be46-1d09b71056fd",
            "content": "Simple outbound message",
            "delivered": False,
            "attempts": 1,
            "metadata": {}
        }
        outbound = Outbound.objects.create(**outbound_message)
        self._restore_post_save_hooks_outbound()  # let tests fire tasks
        return str(outbound.id)

    def make_voice_outbound(self):
        self._replace_post_save_hooks_outbound()  # don't let fixtures fire
        outbound_message = {
            "to_addr": "+27123",
            "vumi_message_id": "075a32da-e1e4-4424-be46-1d09b71056fd",
            "content": "Simple outbound message",
            "delivered": False,
            "metadata": {"voice_speech_url": "http://test.com"}
        }
        outbound = Outbound.objects.create(**outbound_message)
        self._restore_post_save_hooks_outbound()  # let tests fire tasks
        return outbound

    def make_inbound(self, in_reply_to):
        inbound_message = {
            "message_id": str(uuid.uuid4()),
            "in_reply_to": in_reply_to,
            "to_addr": "+27123",
            "from_addr": "020",
            "content": "Call delivered",
            "transport_name": "test_voice",
            "transport_type": "voice",
            "helper_metadata": {}
        }
        inbound = Inbound.objects.create(**inbound_message)
        return str(inbound.id)

    def _replace_get_metric_client(self, session=None):
        return MetricsApiClient(
            auth_token=settings.METRICS_AUTH_TOKEN,
            api_url=settings.METRICS_URL,
            session=self.session)

    def _restore_get_metric_client(self, session=None):
        return MetricsApiClient(
            auth_token=settings.METRICS_AUTH_TOKEN,
            api_url=settings.METRICS_URL,
            session=session)

    def _replace_post_save_hooks_outbound(self):
        post_save.disconnect(psh_fire_msg_action_if_new, sender=Outbound)

    def _replace_post_save_hooks_inbound(self):
        post_save.disconnect(psh_fire_metrics_if_new, sender=Inbound)

    def _restore_post_save_hooks_outbound(self):
        post_save.connect(psh_fire_msg_action_if_new, sender=Outbound)

    def _restore_post_save_hooks_inbound(self):
        post_save.connect(psh_fire_metrics_if_new, sender=Inbound)

    def check_request(
            self, request, method, params=None, data=None, headers=None):
        self.assertEqual(request.method, method)
        if params is not None:
            url = urlparse.urlparse(request.url)
            qs = urlparse.parse_qsl(url.query)
            self.assertEqual(dict(qs), params)
        if headers is not None:
            for key, value in headers.items():
                self.assertEqual(request.headers[key], value)
        if data is None:
            self.assertEqual(request.body, None)
        else:
            self.assertEqual(json.loads(request.body), data)

    def _mount_session(self):
        response = [{
            'name': 'foo',
            'value': 9000,
            'aggregator': 'bar',
        }]
        adapter = RecordingAdapter(json.dumps(response).encode('utf-8'))
        self.session.mount(
            "http://metrics-url/metrics/", adapter)
        return adapter

    def setUp(self):
        super(AuthenticatedAPITestCase, self).setUp()
        self._replace_post_save_hooks_inbound
        tasks.get_metric_client = self._replace_get_metric_client
        self.adapter = self._mount_session()

        self.username = 'testuser'
        self.password = 'testpass'
        self.user = User.objects.create_user(self.username,
                                             'testuser@example.com',
                                             self.password)
        token = Token.objects.create(user=self.user)
        self.token = token.key
        self.client.credentials(HTTP_AUTHORIZATION='Token ' + self.token)
        self.superuser = User.objects.create_superuser('testsu',
                                                       'su@example.com',
                                                       'dummypwd')
        sutoken = Token.objects.create(user=self.superuser)
        self.adminclient.credentials(
            HTTP_AUTHORIZATION='Token %s' % sutoken)

        self.handler = RecordingHandler()
        logger = logging.getLogger('go_http.test')
        logger.setLevel(logging.INFO)
        logger.addHandler(self.handler)

    def tearDown(self):
        self._restore_post_save_hooks_inbound()
        tasks.get_metric_client = self._restore_get_metric_client

    def check_logs(self, msg):
        if self.handler.logs is None:  # nothing to check
            return False
        if type(self.handler.logs) != list:
            [logs] = self.handler.logs
        else:
            logs = self.handler.logs
        for log in logs:
            logline = log.msg.replace("u'", "'")
            if logline == msg:
                return True
        return False


class TestVumiMessagesAPI(AuthenticatedAPITestCase):

    def test_create_outbound_data(self):
        post_outbound = {
            "to_addr": "+27123",
            "vumi_message_id": "075a32da-e1e4-4424-be46-1d09b71056fd",
            "content": "Say something",
            "delivered": False,
            "attempts": 0,
            "metadata": {}
        }
        response = self.client.post('/api/v1/outbound/',
                                    json.dumps(post_outbound),
                                    content_type='application/json')
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        d = Outbound.objects.last()
        self.assertIsNotNone(d.id)
        self.assertEqual(d.version, 1)
        self.assertEqual(str(d.to_addr), "+27123")
        self.assertEqual(d.content, "Say something")
        self.assertEqual(d.delivered, False)
        self.assertEqual(d.attempts, 1)
        self.assertEqual(d.metadata, {})

    def test_create_outbound_data_simple(self):
        post_outbound = {
            "to_addr": "+27123",
            "delivered": "false",
            "metadata": {
                "voice_speech_url": "https://foo.com/file.mp3"
            }
        }
        response = self.client.post('/api/v1/outbound/',
                                    json.dumps(post_outbound),
                                    content_type='application/json')
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        d = Outbound.objects.last()
        self.assertIsNotNone(d.id)
        self.assertEqual(d.version, 1)
        self.assertEqual(str(d.to_addr), "+27123")
        self.assertEqual(d.delivered, False)
        self.assertEqual(d.attempts, 1)
        self.assertEqual(d.metadata, {
            "voice_speech_url": "https://foo.com/file.mp3"
        })

    def test_update_outbound_data(self):
        existing = self.make_outbound()
        patch_outbound = {
            "delivered": "true",
            "attempts": 2
        }
        response = self.client.patch('/api/v1/outbound/%s/' %
                                     existing,
                                     json.dumps(patch_outbound),
                                     content_type='application/json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        d = Outbound.objects.get(pk=existing)
        self.assertEqual(d.version, 1)
        self.assertEqual(str(d.to_addr), "+27123")
        self.assertEqual(d.delivered, True)
        self.assertEqual(d.attempts, 2)

    def test_delete_outbound_data(self):
        existing = self.make_outbound()
        response = self.client.delete('/api/v1/outbound/%s/' %
                                      existing,
                                      content_type='application/json')
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        d = Outbound.objects.filter(id=existing).count()
        self.assertEqual(d, 0)

    def test_created_at_filter_outbound_exists(self):
        existing = Outbound.objects.get(pk=self.make_outbound())
        response = self.client.get('/api/v1/outbound/?%s' % (urlencode({
            'before': (existing.created_at + timedelta(days=1)).isoformat(),
            'after': (existing.created_at - timedelta(days=1)).isoformat(),
        })))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data["count"], 1)
        self.assertEqual(response.data["results"][0]["id"], str(existing.id))

    def test_created_at_filter_outbound_not_exists(self):
        existing = Outbound.objects.get(pk=self.make_outbound())
        response = self.client.get('/api/v1/outbound/?%s' % (urlencode({
            'before': (existing.created_at - timedelta(days=1)).isoformat(),
            'after': (existing.created_at + timedelta(days=1)).isoformat(),
        })))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data["count"], 0)

    def test_create_inbound_data(self):
        existing_outbound = self.make_outbound()
        out = Outbound.objects.get(pk=existing_outbound)
        message_id = str(uuid.uuid4())
        post_inbound = {
            "message_id": message_id,
            "in_reply_to": out.vumi_message_id,
            "to_addr": "+27123",
            "from_addr": "020",
            "content": "Call delivered",
            "transport_name": "test_voice",
            "transport_type": "voice",
            "helper_metadata": {}
        }
        with patch.object(ConcurrencyLimiter, 'decr_message_count') as \
                mock_method:
            response = self.client.post('/api/v1/inbound/',
                                        json.dumps(post_inbound),
                                        content_type='application/json')
            mock_method.assert_not_called()
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        d = Inbound.objects.last()
        self.assertIsNotNone(d.id)
        self.assertEqual(d.message_id, message_id)
        self.assertEqual(d.to_addr, "+27123")
        self.assertEqual(d.from_addr, "020")
        self.assertEqual(d.content, "Call delivered")
        self.assertEqual(d.transport_name, "test_voice")
        self.assertEqual(d.transport_type, "voice")
        self.assertEqual(d.helper_metadata, {})

    def test_update_inbound_data(self):
        existing_outbound = self.make_outbound()
        out = Outbound.objects.get(pk=existing_outbound)
        existing = self.make_inbound(out.vumi_message_id)

        patch_inbound = {
            "content": "Opt out"
        }
        response = self.client.patch('/api/v1/inbound/%s/' %
                                     existing,
                                     json.dumps(patch_inbound),
                                     content_type='application/json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        d = Inbound.objects.get(pk=existing)
        self.assertEqual(d.to_addr, "+27123")
        self.assertEqual(d.from_addr, "020")
        self.assertEqual(d.content, "Opt out")
        self.assertEqual(d.transport_name, "test_voice")
        self.assertEqual(d.transport_type, "voice")
        self.assertEqual(d.helper_metadata, {})

    def test_delete_inbound_data(self):
        existing_outbound = self.make_outbound()
        out = Outbound.objects.get(pk=existing_outbound)
        existing = self.make_inbound(out.vumi_message_id)
        response = self.client.delete('/api/v1/inbound/%s/' %
                                      existing,
                                      content_type='application/json')
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        d = Inbound.objects.filter(id=existing).count()
        self.assertEqual(d, 0)

    def test_create_inbound_event_message(self):
        existing_outbound = self.make_outbound()
        out = Outbound.objects.get(pk=existing_outbound)
        out.last_sent_time = out.created_at
        out.save()
        message_id = str(uuid.uuid4())
        post_inbound = {
            "message_id": message_id,
            "in_reply_to": out.vumi_message_id,
            "to_addr": "0.0.0.0:9001",
            "from_addr": "020",
            "content": "Call delivered",
            "transport_name": "test_voice",
            "transport_type": "voice",
            "helper_metadata": {},
            "session_event": "close"
        }

        with patch.object(ConcurrencyLimiter, 'decr_message_count') as \
                mock_method:
            response = self.client.post('/api/v1/inbound/',
                                        json.dumps(post_inbound),
                                        content_type='application/json')
            mock_method.assert_called_once_with("text", out.created_at)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        d = Inbound.objects.last()
        self.assertIsNotNone(d.id)
        self.assertEqual(d.message_id, message_id)
        self.assertEqual(d.to_addr, "0.0.0.0:9001")
        self.assertEqual(d.from_addr, "020")
        self.assertEqual(d.content, "Call delivered")
        self.assertEqual(d.transport_name, "test_voice")
        self.assertEqual(d.transport_type, "voice")
        self.assertEqual(d.helper_metadata, {"session_event": "close"})

    def test_event_ack(self):
        existing = self.make_outbound()

        d = Outbound.objects.get(pk=existing)
        ack = {
            "message_type": "event",
            "event_id": "b04ec322fc1c4819bc3f28e6e0c69de6",
            "event_type": "ack",
            "user_message_id": d.vumi_message_id,
            "helper_metadata": {},
            "timestamp": "2015-10-28 16:19:37.485612",
            "sent_message_id": "external-id"
        }
        response = self.client.post('/api/v1/events',
                                    json.dumps(ack),
                                    content_type='application/json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        d = Outbound.objects.get(pk=existing)
        self.assertEqual(d.delivered, True)
        self.assertEqual(d.attempts, 1)
        self.assertEqual(d.metadata["ack_timestamp"],
                         "2015-10-28 16:19:37.485612")
        self.assertEquals(False, self.check_logs(
            "Message: 'Simple outbound message' sent to '+27123'"))

    def test_event_delivery_report(self):
        existing = self.make_outbound()
        d = Outbound.objects.get(pk=existing)
        dr = {
            "message_type": "event",
            "event_id": "b04ec322fc1c4819bc3f28e6e0c69de6",
            "event_type": "delivery_report",
            "user_message_id": d.vumi_message_id,
            "helper_metadata": {},
            "timestamp": "2015-10-28 16:20:37.485612",
            "sent_message_id": "external-id"
        }
        response = self.client.post('/api/v1/events',
                                    json.dumps(dr),
                                    content_type='application/json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        d = Outbound.objects.get(pk=existing)
        self.assertEqual(d.delivered, True)
        self.assertEqual(d.attempts, 1)
        self.assertEqual(d.metadata["delivery_timestamp"],
                         "2015-10-28 16:20:37.485612")
        self.assertEquals(False, self.check_logs(
            "Message: 'Simple outbound message' sent to '+27123'"))

    def test_event_nack_first(self):
        existing = self.make_outbound()
        d = Outbound.objects.get(pk=existing)
        post_save.connect(psh_fire_msg_action_if_new, sender=Outbound)
        nack = {
            "message_type": "event",
            "event_id": "b04ec322fc1c4819bc3f28e6e0c69de6",
            "event_type": "nack",
            "nack_reason": "no answer",
            "user_message_id": d.vumi_message_id,
            "helper_metadata": {},
            "timestamp": "2015-10-28 16:20:37.485612",
            "sent_message_id": "external-id"
        }
        response = self.client.post('/api/v1/events',
                                    json.dumps(nack),
                                    content_type='application/json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        c = Outbound.objects.get(pk=existing)
        self.assertEqual(c.delivered, False)
        self.assertEqual(c.attempts, 2)
        self.assertEqual(c.metadata["nack_reason"],
                         "no answer")
        self.assertEquals(True, self.check_logs(
            "Message: 'Simple outbound message' sent to '+27123' "
            "[session_event: new]"))
        # TODO: Bring metrics back
        # self.assertEquals(
        #     True,
        #     self.check_logs("Metric: 'vumimessage.tries' [sum] -> 1"))

    def test_event_nack_last(self):
        # Be assured this is last message attempt
        outbound_message = {
            "to_addr": "+27123",
            "vumi_message_id": "08b34de7-c6da-4853-a74d-9458533ed169",
            "content": "Simple outbound message",
            "delivered": False,
            "attempts": 3,
            "metadata": {}
        }
        failed = Outbound.objects.create(**outbound_message)
        failed.last_sent_time = failed.created_at
        failed.save()
        post_save.connect(psh_fire_msg_action_if_new, sender=Outbound)
        nack = {
            "message_type": "event",
            "event_id": "b04ec322fc1c4819bc3f28e6e0c69de6",
            "event_type": "nack",
            "nack_reason": "no answer",
            "user_message_id": failed.vumi_message_id,
            "helper_metadata": {},
            "timestamp": "2015-10-28 16:20:37.485612",
            "sent_message_id": "external-id"
        }
        response = self.client.post('/api/v1/events',
                                    json.dumps(nack),
                                    content_type='application/json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        d = Outbound.objects.get(pk=failed.id)
        self.assertEqual(d.delivered, False)
        self.assertEqual(d.attempts, 3)  # not moved on as last attempt passed
        self.assertEqual(d.metadata["nack_reason"],
                         "no answer")
        self.assertEquals(False, self.check_logs(
            "Message: 'Simple outbound message' sent to '+27123'"
            "[session_event: new]"))
        # TODO: Bring metrics back
        # self.assertEquals(
        #     False,
        #     self.check_logs("Metric: 'vumimessage.tries' [sum] -> 1"))
        # self.assertEquals(
        #     True,
        #     self.check_logs("Metric: 'vumimessage.maxretries' [sum] -> 1"))


class TestJunebugMessagesAPI(AuthenticatedAPITestCase):
    def test_event_missing_fields(self):
        '''
        If there are missing fields in the request, and error response should
        be returned.
        '''
        response = self.client.post(
            '/api/v1/events/junebug', json.dumps({}),
            content_type='application/json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_event_no_message(self):
        '''
        If we cannot find the message for the event, and error response should
        be returned.
        '''
        ack = {
            "event_type": "submitted",
            "message_id": 'bad-message-id',
            "channel-id": "channel-uuid-1234",
            "timestamp": "2015-10-28 16:19:37.485612",
            "event_details": {},
        }
        response = self.client.post(
            '/api/v1/events/junebug', json.dumps(ack),
            content_type='application/json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_event_ack(self):
        '''A submitted event should update the message object accordingly.'''
        existing = self.make_outbound()

        d = Outbound.objects.get(pk=existing)
        ack = {
            "event_type": "submitted",
            "message_id": d.vumi_message_id,
            "channel-id": "channel-uuid-1234",
            "timestamp": "2015-10-28 16:19:37.485612",
            "event_details": {},
        }

        response = self.client.post(
            '/api/v1/events/junebug', json.dumps(ack),
            content_type='application/json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        d = Outbound.objects.get(pk=existing)
        self.assertEqual(d.delivered, True)
        self.assertEqual(d.attempts, 1)
        self.assertEqual(
            d.metadata["ack_timestamp"], "2015-10-28 16:19:37.485612")
        self.assertEquals(False, self.check_logs(
            "Message: 'Simple outbound message' sent to '+27123'"))

    def test_event_nack(self):
        '''
        A rejected event should retry and update the message object accordingly
        '''
        existing = self.make_outbound()
        d = Outbound.objects.get(pk=existing)
        post_save.connect(psh_fire_msg_action_if_new, sender=Outbound)
        nack = {
            "event_type": "rejected",
            "message_id": d.vumi_message_id,
            "channel-id": "channel-uuid-1234",
            "timestamp": "2015-10-28 16:19:37.485612",
            "event_details": {"reason": "No answer"},
        }
        response = self.client.post(
            '/api/v1/events/junebug', json.dumps(nack),
            content_type='application/json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        c = Outbound.objects.get(pk=existing)
        self.assertEqual(c.delivered, False)
        self.assertEqual(c.attempts, 2)
        self.assertEqual(
            c.metadata["nack_reason"], {"reason": "No answer"})
        self.assertEquals(True, self.check_logs(
            "Message: 'Simple outbound message' sent to '+27123' "
            "[session_event: new]"))

    def test_event_delivery_succeeded(self):
        '''A successful delivery should update the message accordingly.'''
        existing = self.make_outbound()
        d = Outbound.objects.get(pk=existing)
        dr = {
            "event_type": "delivery_succeeded",
            "message_id": d.vumi_message_id,
            "channel-id": "channel-uuid-1234",
            "timestamp": "2015-10-28 16:19:37.485612",
            "event_details": {},
        }
        response = self.client.post(
            '/api/v1/events/junebug', json.dumps(dr),
            content_type='application/json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        d = Outbound.objects.get(pk=existing)
        self.assertEqual(d.delivered, True)
        self.assertEqual(d.attempts, 1)
        self.assertEqual(
            d.metadata["delivery_timestamp"], "2015-10-28 16:19:37.485612")
        self.assertEquals(False, self.check_logs(
            "Message: 'Simple outbound message' sent to '+27123'"))

    def test_event_delivery_failed(self):
        '''
        A failed delivery should retry and update the message accordingly.
        '''
        existing = self.make_outbound()
        d = Outbound.objects.get(pk=existing)
        dr = {
            "event_type": "delivery_failed",
            "message_id": d.vumi_message_id,
            "channel-id": "channel-uuid-1234",
            "timestamp": "2015-10-28 16:19:37.485612",
            "event_details": {},
        }
        response = self.client.post(
            '/api/v1/events/junebug', json.dumps(dr),
            content_type='application/json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        d = Outbound.objects.get(pk=existing)
        self.assertEqual(d.delivered, False)
        self.assertEqual(d.attempts, 2)
        self.assertEqual(
            d.metadata["delivery_failed_reason"], {})
        self.assertEquals(False, self.check_logs(
            "Message: 'Simple outbound message' sent to '+27123'"))

    def test_create_inbound_junebug_message(self):
        existing_outbound = self.make_outbound()
        out = Outbound.objects.get(pk=existing_outbound)
        out.last_sent_time = out.created_at
        out.save()
        message_id = str(uuid.uuid4())
        post_inbound = {
            "message_id": message_id,
            "reply_to": "test_id",
            "to": "0.0.0.0:9001",
            "from": out.to_addr,
            "content": "Call delivered",
            "channel_id": "test_voice",
            "channel_data": {"session_event": "close"}
        }
        with patch.object(ConcurrencyLimiter, 'decr_message_count') as \
                mock_method:
            response = self.client.post('/api/v1/inbound/',
                                        json.dumps(post_inbound),
                                        content_type='application/json')
            mock_method.assert_called_once_with("text", out.created_at)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        d = Inbound.objects.last()
        self.assertIsNotNone(d.id)
        self.assertEqual(d.message_id, message_id)
        self.assertEqual(d.to_addr, "0.0.0.0:9001")
        self.assertEqual(d.from_addr, "+27123")
        self.assertEqual(d.content, "Call delivered")
        self.assertEqual(d.transport_name, "test_voice")
        self.assertEqual(d.transport_type, None)
        self.assertEqual(d.helper_metadata, {"session_event": "close"})


class TestMetricsAPI(AuthenticatedAPITestCase):

    def test_metrics_read(self):
        # Setup
        # Execute
        response = self.client.get('/api/metrics/',
                                   content_type='application/json')
        # Check
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            response.data["metrics_available"], [
                'inbounds.created.sum',
                'vumimessage.tries.sum',
                'vumimessage.maxretries.sum',
                'vumimessage.obd.tries.sum',
                'vumimessage.obd.successful.sum',
                'vumimessage.obd.unsuccessful.sum',
                'message.failures.sum',
                'message.sent.sum',
                'sender.send_message.connection_error.sum',
                'sender.send_message.http_error.400.sum',
                'sender.send_message.http_error.401.sum',
                'sender.send_message.http_error.403.sum',
                'sender.send_message.http_error.404.sum',
                'sender.send_message.http_error.500.sum',
                'sender.send_message.timeout.sum',
            ]
        )

    @responses.activate
    def test_post_metrics(self):
        # Setup
        # deactivate Testsession for this test
        self.session = None
        responses.add(responses.POST,
                      "http://metrics-url/metrics/",
                      json={"foo": "bar"},
                      status=200, content_type='application/json')
        # Execute
        response = self.client.post('/api/metrics/',
                                    content_type='application/json')
        # Check
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(response.data["scheduled_metrics_initiated"], True)


class TestMetrics(AuthenticatedAPITestCase):

    def test_direct_fire(self):
        # Setup
        adapter = self._mount_session()
        # Execute
        result = fire_metric.apply_async(kwargs={
            "metric_name": 'foo.last',
            "metric_value": 1,
            "session": self.session
        })
        # Check
        self.check_request(
            adapter.request, 'POST',
            data={"foo.last": 1.0}
        )
        self.assertEqual(result.get(),
                         "Fired metric <foo.last> with value <1.0>")

    def test_created_metrics(self):
        # Setup
        adapter = self._mount_session()
        # reconnect metric post_save hook
        post_save.connect(psh_fire_metrics_if_new, sender=Inbound)
        # make outbound
        existing_outbound = self.make_outbound()
        out = Outbound.objects.get(pk=existing_outbound)

        # Execute
        self.make_inbound(out.vumi_message_id)

        # Check
        self.check_request(
            adapter.request, 'POST',
            data={"inbounds.created.sum": 1.0}
        )
        # remove post_save hooks to prevent teardown errors
        post_save.disconnect(psh_fire_metrics_if_new, sender=Inbound)


class TestHealthcheckAPI(AuthenticatedAPITestCase):

    def test_healthcheck_read(self):
        # Setup
        # Execute
        response = self.client.get('/api/health/',
                                   content_type='application/json')
        # Check
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data["up"], True)
        self.assertEqual(response.data["result"]["database"], "Accessible")


class TestUserCreation(AuthenticatedAPITestCase):

    def test_create_user_and_token(self):
        # Setup
        user_request = {"email": "test@example.org"}
        # Execute
        request = self.adminclient.post('/api/v1/user/token/', user_request)
        token = request.json().get('token', None)
        # Check
        self.assertIsNotNone(
            token, "Could not receive authentication token on post.")
        self.assertEqual(
            request.status_code, 201,
            "Status code on /api/v1/user/token/ was %s (should be 201)."
            % request.status_code)

    def test_create_user_and_token_fail_nonadmin(self):
        # Setup
        user_request = {"email": "test@example.org"}
        # Execute
        request = self.client.post('/api/v1/user/token/', user_request)
        error = request.json().get('detail', None)
        # Check
        self.assertIsNotNone(
            error, "Could not receive error on post.")
        self.assertEqual(
            error, "You do not have permission to perform this action.",
            "Error message was unexpected: %s."
            % error)

    def test_create_user_and_token_not_created(self):
        # Setup
        user_request = {"email": "test@example.org"}
        # Execute
        request = self.adminclient.post('/api/v1/user/token/', user_request)
        token = request.json().get('token', None)
        # And again, to get the same token
        request2 = self.adminclient.post('/api/v1/user/token/', user_request)
        token2 = request2.json().get('token', None)

        # Check
        self.assertEqual(
            token, token2,
            "Tokens are not equal, should be the same as not recreated.")

    def test_create_user_new_token_nonadmin(self):
        # Setup
        user_request = {"email": "test@example.org"}
        request = self.adminclient.post('/api/v1/user/token/', user_request)
        token = request.json().get('token', None)
        cleanclient = APIClient()
        cleanclient.credentials(HTTP_AUTHORIZATION='Token %s' % token)
        # Execute
        request = cleanclient.post('/api/v1/user/token/', user_request)
        error = request.json().get('detail', None)
        # Check
        # new user should not be admin
        self.assertIsNotNone(
            error, "Could not receive error on post.")
        self.assertEqual(
            error, "You do not have permission to perform this action.",
            "Error message was unexpected: %s."
            % error)


class TestFormatter(TestCase):

    @override_settings(
        VOICE_TO_ADDR_FORMATTER='message_sender.formatters.noop')
    def test_noop(self):
        cb = load_callable(settings.VOICE_TO_ADDR_FORMATTER)
        self.assertEqual(cb('12345'), '12345')

    @override_settings(
        VOICE_TO_ADDR_FORMATTER='message_sender.formatters.vas2nets_voice')
    def test_vas2nets_voice(self):
        cb = load_callable(settings.VOICE_TO_ADDR_FORMATTER)
        self.assertEqual(cb('+23456'), '9056')
        self.assertEqual(cb('23456'), '9056')

    @override_settings(
        VOICE_TO_ADDR_FORMATTER='message_sender.formatters.vas2nets_text')
    def test_vas2nets_text(self):
        cb = load_callable(settings.VOICE_TO_ADDR_FORMATTER)
        self.assertEqual(cb('+23456'), '23456')
        self.assertEqual(cb('23456'), '23456')


class TestFactory(TestCase):

    @override_settings(MESSAGE_BACKEND_TEXT='junebug',
                       JUNEBUG_API_URL_TEXT='http://example.com/',
                       JUNEBUG_API_AUTH_TEXT=('username', 'password'))
    def test_create_junebug_text(self):
        message_sender = MessageClientFactory.create('text')
        self.assertTrue(isinstance(message_sender, JunebugApiSender))
        self.assertEqual(message_sender.api_url, 'http://example.com/')
        self.assertEqual(message_sender.auth, ('username', 'password'))

    @override_settings(MESSAGE_BACKEND_VOICE='junebug',
                       JUNEBUG_API_URL_VOICE='http://example.com/voice',
                       JUNEBUG_API_AUTH_VOICE=('username', 'password'))
    def test_create_junebug_voice(self):
        message_sender = MessageClientFactory.create('voice')
        self.assertTrue(isinstance(message_sender, JunebugApiSender))
        self.assertEqual(message_sender.api_url, 'http://example.com/voice')
        self.assertEqual(message_sender.auth, ('username', 'password'))

    @override_settings(MESSAGE_BACKEND_TEXT='vumi',
                       VUMI_CONVERSATION_KEY_TEXT='conv-key',
                       VUMI_ACCOUNT_KEY_TEXT='account-key',
                       VUMI_ACCOUNT_TOKEN_TEXT='account-token',
                       VUMI_API_URL_TEXT='http://example.com/')
    def test_create_vumi_text(self):
        message_sender = MessageClientFactory.create('text')
        self.assertTrue(isinstance(message_sender, HttpApiSender))
        self.assertEqual(
            message_sender.api_url, 'http://example.com/')
        self.assertEqual(message_sender.account_key, 'account-key')
        self.assertEqual(message_sender.conversation_key, 'conv-key')
        self.assertEqual(message_sender.conversation_token, 'account-token')

    @override_settings(MESSAGE_BACKEND_VOICE='vumi',
                       VUMI_CONVERSATION_KEY_VOICE='conv-key',
                       VUMI_ACCOUNT_KEY_VOICE='account-key',
                       VUMI_ACCOUNT_TOKEN_VOICE='account-token',
                       VUMI_API_URL_VOICE='http://example.com/')
    def test_create_vumi_voice(self):
        message_sender = MessageClientFactory.create('voice')
        self.assertTrue(isinstance(message_sender, HttpApiSender))
        self.assertEqual(
            message_sender.api_url, 'http://example.com/')
        self.assertEqual(message_sender.account_key, 'account-key')
        self.assertEqual(message_sender.conversation_key, 'conv-key')
        self.assertEqual(message_sender.conversation_token, 'account-token')

    @override_settings(MESSAGE_BACKEND_VOICE='unknown')
    def test_create_unknown(self):
        '''
        The message client factory should raise an exception if an unknown
        message type is specified.
        '''
        self.assertRaises(
            FactoryException, MessageClientFactory.create, 'voice')

    @override_settings(MESSAGE_BACKEND_VOICE=None)
    def test_create_no_backend_type_specified(self):
        '''
        If no message backend is specified, an error should be raised when
        getting the message client.
        '''
        self.assertRaises(
            FactoryException, MessageClientFactory.create, 'voice')


class TestJunebugAPISender(TestCase):
    @override_settings(MESSAGE_BACKEND_TEXT='junebug',
                       JUNEBUG_API_URL_TEXT='http://example.com/',
                       JUNEBUG_API_FROM_TEXT='+4321')
    @responses.activate
    def test_send_text(self):
        '''
        Using the send_text function should send a request to Junebug with the
        correct JSON data.
        '''
        responses.add(
            responses.POST, "http://example.com/",
            json={"result": {"message_id": "message-uuid"}}, status=200,
            content_type='application/json')

        message_sender = MessageClientFactory.create('text')
        res = message_sender.send_text('+1234', 'Test', session_event='resume')

        self.assertEqual(res['message_id'], 'message-uuid')

        [r] = responses.calls
        r = json.loads(r.request.body)
        self.assertEqual(r['to'], '+1234')
        self.assertEqual(r['from'], '+4321')
        self.assertEqual(r['content'], 'Test')
        self.assertEqual(r['channel_data']['session_event'], 'resume')
        self.assertEqual(
            r['event_url'], 'http://example.com/api/v1/events/junebug')

    @override_settings(MESSAGE_BACKEND_VOICE='junebug',
                       JUNEBUG_API_URL_VOICE='http://example.com/',
                       JUNEBUG_API_FROM_VOICE='+4321')
    @responses.activate
    def test_send_voice(self):
        '''
        Using the send_voice function should send a request to Junebug with the
        correct JSON data.
        '''
        responses.add(
            responses.POST, "http://example.com/",
            json={"result": {"message_id": "message-uuid"}}, status=200,
            content_type='application/json')

        message_sender = MessageClientFactory.create('voice')
        res = message_sender.send_voice(
            '+1234', 'Test', speech_url='http://test.mp3', wait_for='#',
            session_event='resume')

        self.assertEqual(res['message_id'], 'message-uuid')

        [r] = responses.calls
        r = json.loads(r.request.body)
        self.assertEqual(r['to'], '+1234')
        self.assertEqual(r['from'], '+4321')
        self.assertEqual(r['content'], 'Test')
        self.assertEqual(r['channel_data']['session_event'], 'resume')
        self.assertEqual(
            r['channel_data']['voice']['speech_url'], 'http://test.mp3')
        self.assertEqual(r['channel_data']['voice']['wait_for'], '#')
        self.assertEqual(
            r['event_url'], 'http://example.com/api/v1/events/junebug')

    @override_settings(MESSAGE_BACKEND_VOICE='junebug')
    def test_fire_metric(self):
        '''
        Using the fire_metric function should result in an exception being
        raised, since Junebug doesn't support metrics sending.
        '''
        message_sender = MessageClientFactory.create('voice')
        self.assertRaises(
            JunebugApiSenderException, message_sender.fire_metric, 'foo.bar',
            3.0, agg='sum')


class TestConcurrencyLimiter(AuthenticatedAPITestCase):
    def make_outbound(self, to_addr):
        self._replace_post_save_hooks_outbound()  # don't let fixtures fire
        outbound_message = {
            "to_addr": to_addr,
            "vumi_message_id": "075a32da-e1e4-4424-be46-1d09b71056fd",
            "content": "Simple outbound message",
            "delivered": False,
            "metadata": {"voice_speech_url": "http://test.com"}
        }
        outbound = Outbound.objects.create(**outbound_message)
        self._restore_post_save_hooks_outbound()  # let tests fire tasks
        return outbound

    def set_cache_entry(self, msg_type, bucket, value):
        key = "%s_messages_at_%s" % (msg_type, bucket)
        self.fake_cache.cache_data[key] = value

    def setUp(self):
        super(TestConcurrencyLimiter, self).setUp()
        self.fake_cache = MockCache()

    @override_settings(CONCURRENT_VOICE_LIMIT=2, VOICE_MESSAGE_DELAY=10,
                       VOICE_MESSAGE_TIMEOUT=20)
    @patch('time.time', MagicMock(return_value=1479131658.000000))
    @patch('django.core.cache.cache.get')
    @patch('django.core.cache.cache.add')
    @patch('django.core.cache.cache.incr')
    def test_limiter_limit_not_reached(self, mock_incr, mock_add, mock_get):
        """
        Messages under the limit should get sent.
        """
        # Fake cache calls
        mock_incr.side_effect = self.fake_cache.incr
        mock_add.side_effect = self.fake_cache.add
        mock_get.side_effect = self.fake_cache.get

        outbound1 = self.make_outbound(to_addr="+27123")
        outbound2 = self.make_outbound(to_addr="+27987")

        send_message(outbound1.pk)
        send_message(outbound2.pk)

        self.assertTrue(self.check_logs(
            "Message: None sent to '%s' [session_event: new]" % (
                outbound1.to_addr)))
        self.assertTrue(self.check_logs(
            "Message: None sent to '%s' [session_event: new]" % (
                outbound2.to_addr)))
        outbound1.refresh_from_db()
        self.assertIsNotNone(outbound1.last_sent_time)
        outbound2.refresh_from_db()
        self.assertIsNotNone(outbound2.last_sent_time)
        self.assertEqual(len(self.fake_cache.cache_data), 1)
        bucket = 1479131658 // 60  # time() // bucket_size
        self.assertEqual(
            self.fake_cache.cache_data["voice_messages_at_%s" % bucket], 2)

    @override_settings(CONCURRENT_VOICE_LIMIT=1, VOICE_MESSAGE_DELAY=10,
                       VOICE_MESSAGE_TIMEOUT=20)
    @patch('time.time', MagicMock(return_value=1479131658.000000))
    @patch('django.core.cache.cache.get')
    @patch('django.core.cache.cache.add')
    @patch('django.core.cache.cache.incr')
    @patch('message_sender.tasks.send_message.retry')
    def test_limiter_limit_reached(self, mock_retry, mock_incr, mock_add,
                                   mock_get):
        """
        Messages under the limit should get sent. Messages over the limit
        should get retried
        """
        mock_retry.side_effect = Retry

        # Fake cache calls
        mock_incr.side_effect = self.fake_cache.incr
        mock_add.side_effect = self.fake_cache.add
        mock_get.side_effect = self.fake_cache.get

        outbound1 = self.make_outbound(to_addr="+27123")
        outbound2 = self.make_outbound(to_addr="+27987")

        send_message(outbound1.pk)
        with self.assertRaises(Retry):
            send_message(outbound2.pk)
        mock_retry.assert_called_with(countdown=10)

        self.assertTrue(self.check_logs(
            "Message: None sent to '%s' [session_event: new]" % (
                outbound1.to_addr)))
        self.assertFalse(self.check_logs(
            "Message: None sent to '%s' [session_event: new]" % (
                outbound2.to_addr)))
        outbound1.refresh_from_db()
        self.assertIsNotNone(outbound1.last_sent_time)
        outbound2.refresh_from_db()
        self.assertIsNone(outbound2.last_sent_time)
        self.assertEqual(len(self.fake_cache.cache_data), 1)
        bucket = 1479131658 // 60  # time() // bucket_size
        self.assertEqual(
            self.fake_cache.cache_data["voice_messages_at_%s" % bucket], 1)

    @override_settings(VOICE_MESSAGE_DELAY=100, VOICE_MESSAGE_TIMEOUT=120)
    @patch('time.time', MagicMock(return_value=1479131640.000000))
    @patch('django.core.cache.cache.get')
    def test_limiter_buckets(self, mock_get):
        """
        The correct buckets should count towards the message count.
        """

        # Fake cache calls
        mock_get.side_effect = self.fake_cache.get
        now = 1479131640

        self.set_cache_entry("voice", (now - 200) // 60, 1)  # Too old
        self.set_cache_entry("voice", (now - 121) // 60, 10)  # Over delay
        self.set_cache_entry("voice", (now - 120) // 60, 100)  # Within delay
        self.set_cache_entry("voice", now // 60, 1000)  # Now
        self.set_cache_entry("voice", (now + 60) // 60, 10000)  # In future

        count = ConcurrencyLimiter.get_current_message_count("voice", 120)
        self.assertEqual(count, 1100)

    @override_settings(CONCURRENT_VOICE_LIMIT=1,
                       VOICE_MESSAGE_DELAY=100, VOICE_MESSAGE_TIMEOUT=120)
    @patch('time.time', MagicMock(return_value=1479131658.000000))
    @patch('django.core.cache.cache.get_or_set')
    @patch('django.core.cache.cache.decr')
    def test_limiter_decr_count(self, mock_decr, mock_get_or_set):
        """
        Events for messages should decrement the counter unless the message is
        too old.
        """

        # Fake cache calls
        mock_get_or_set.side_effect = self.fake_cache.get_or_set
        mock_decr.side_effect = self.fake_cache.decr

        self.set_cache_entry("voice", 1479131535 // 60, 1)  # Past delay
        self.set_cache_entry("voice", 1479131588 // 60, 1)  # Within delay
        self.set_cache_entry("voice", 1479131648 // 60, -0)  # Invalid value

        ConcurrencyLimiter.decr_message_count(
            "voice", datetime.fromtimestamp(1479131535))
        ConcurrencyLimiter.decr_message_count(
            "voice", datetime.fromtimestamp(1479131588))
        ConcurrencyLimiter.decr_message_count(
            "voice", datetime.fromtimestamp(1479131608))

        self.assertEqual(self.fake_cache.cache_data, {
            "voice_messages_at_24652192": 1, "voice_messages_at_24652193": 0,
            "voice_messages_at_24652194": 0})


class TestRequeueFailedTasks(AuthenticatedAPITestCase):
    def make_outbound(self, to_addr):
        self._replace_post_save_hooks_outbound()  # don't let fixtures fire
        outbound_message = {
            "to_addr": to_addr,
            "vumi_message_id": "075a32da-e1e4-4424-be46-1d09b71056fd",
            "content": "Simple outbound message",
            "delivered": False,
            "metadata": {"voice_speech_url": "http://test.com"}
        }
        outbound = Outbound.objects.create(**outbound_message)
        self._restore_post_save_hooks_outbound()  # let tests fire tasks
        return outbound

    def test_requeue(self):
        outbound1 = self.make_outbound(to_addr="+27123")
        outbound2 = self.make_outbound(to_addr="+27987")
        OutboundSendFailure.objects.create(
            outbound=outbound1,
            task_id=uuid.uuid4(),
            initiated_at=timezone.now(),
            reason='Error')

        requeue_failed_tasks()

        outbound1.refresh_from_db()
        self.assertIsNotNone(outbound1.last_sent_time)
        outbound2.refresh_from_db()
        self.assertIsNone(outbound2.last_sent_time)
        self.assertEqual(OutboundSendFailure.objects.all().count(), 0)


class TestFailedTaskAPI(AuthenticatedAPITestCase):

    def make_outbound(self, to_addr):
        self._replace_post_save_hooks_outbound()  # don't let fixtures fire
        outbound_message = {
            "to_addr": to_addr,
            "vumi_message_id": "075a32da-e1e4-4424-be46-1d09b71056fd",
            "content": "Simple outbound message",
            "delivered": False,
            "metadata": {"voice_speech_url": "http://test.com"}
        }
        outbound = Outbound.objects.create(**outbound_message)
        self._restore_post_save_hooks_outbound()  # let tests fire tasks
        return outbound

    @responses.activate
    def test_failed_tasks_requeue(self):
        outbound1 = self.make_outbound(to_addr="+27123")
        OutboundSendFailure.objects.create(
            outbound=outbound1,
            task_id=uuid.uuid4(),
            initiated_at=timezone.now(),
            reason='Error')

        response = self.client.post('/api/v1/failed-tasks/',
                                    content_type='application/json')
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(response.data["requeued_failed_tasks"], True)
        self.assertEqual(OutboundSendFailure.objects.all().count(), 0)
